/*
 * config.c: URL filter daemon
 *
 * Copyright (C) 2021 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <uci.h>

#include <syslog.h>

#include "config.h"
#include "filter.h"

static struct uci_context *url_ctx;
static struct uci_package *url_pkg;

profile *url_profile;
filter *url_filter;
global url_global;
dns_cache *global_dns_cache;

size_t profile_cnt = 0;
size_t filter_cnt = 0;
size_t dns_cache_cnt = 0;

// This function used to store the name of profile in the glabal
// array url_profile which is identified with i index
// @param i is the index of glabal array url_profile, this url_profile
//          for which the name of profile will be stored
// @param s is the pointer to the uci structure to be used to extract name of
//          profile entry in the profile section
// @return nothing
static void get_profile_name(size_t i, struct uci_section *s)
{
	struct uci_element *element = NULL;
	struct uci_option *option = NULL;

	(url_profile + i)->name[0] = 0;

	uci_foreach_element(&s->options, element) {
		option = uci_to_option(element);

		if (strcmp(option->e.name, "name") == 0) {
			strncpy((url_profile + i)->name, option->v.string, sizeof((url_profile + i)->name));
			break;
		}
	}
}

// This function returns the number of list elements of name option_name in uci
// section s
// @param s is pointer to uci section
// @param option_name is name of list type
// @return no of entries on success, 0 otherwise
static unsigned int get_list_entries_count(struct uci_section *s, char *option_name)
{
	struct uci_element *element = NULL, *bl_element = NULL;
	struct uci_option *option;
	char *tok, *tok_r;
	unsigned int count = 0;
	char *local_str;

	uci_foreach_element(&s->options, element) {
		option = (uci_to_option(element));
		if (strcmp(option->e.name, option_name) == 0) {
			switch (option->type) {
			case UCI_TYPE_LIST:
				// handle list available as
				// list name value1
				// list name value2
				// ....
				uci_foreach_element(&option->v.list, bl_element) {
					count++;
				}
				break;
			case UCI_TYPE_STRING:
				// handle list available as
				// option name value1 value2 value3 ...
				if (!option->v.string || (option->v.string)[0] == '\0')
					break;

				local_str = strdup(option->v.string);
				tok = strtok_r(local_str, " ", &tok_r);
				while (tok != NULL) {
					count++;
					tok = strtok_r(NULL, " ", &tok_r);
				}
				free(local_str);
				break;
			default:
				return 0;
			}
		}
	}

	return count;
}

// This function allocated memory to url_profile->blacklist_url and
// url_profile->whitelist_url variables. It also stores the value of
// the number of urls stored in this profile that is url_profile->blacklist_url_cnt
// and url_profile->blacklist_url_cnt
// @param i is the index of  glabal array url_profile, this the url_profile
//          for which the memory allocation will be done
// @param s is the pointer to the uci structure to be used to extract number of
//          blacklist url and whitelist url entries in the profile section
// @return nothing
static void create_url_list(size_t i, struct uci_section *s)
{
	unsigned int blacklist_url_count = get_list_entries_count(s, "blacklist_url");
	unsigned int whitelist_url_count = get_list_entries_count(s, "whitelist_url");

	// Allocate memory to array of strings, this will define the number
	// of entries in the variable, so no of array elements
	(url_profile + i)->blacklist_url = (char **)calloc(blacklist_url_count, sizeof(char *));

	(url_profile + i)->whitelist_url = (char **)calloc(whitelist_url_count, sizeof(char *));

	(url_profile + i)->blacklist_url_cnt = blacklist_url_count;
	(url_profile + i)->whitelist_url_cnt = whitelist_url_count;

	int j;

	for (j = 0; j < blacklist_url_count; j++)
		(url_profile + i)->blacklist_url[j] = (char *)calloc(MAX_URL_LENGTH, sizeof(char));

	for (j = 0; j < whitelist_url_count; j++)
		(url_profile + i)->whitelist_url[j] = (char *)calloc(MAX_URL_LENGTH, sizeof(char));
}

// this function read the values in the url list in uci file (called once for each
// blacklist and whitelist) and stores this value in corresponding string
// @param s is pointer to uci section
// @param option_name is name of list type parameter to be read
// @param value is pointer to array of strings in which value is to be stored
// @return nothing
static void populate_string_from_list(struct uci_section *s, char *option_name, char **value)
{
	struct uci_element *element = NULL;
	struct uci_element *bl_element = NULL;
	struct uci_option *option;
	size_t len;
	char *tok, *tok_r;
	unsigned int i = 0;
	char *local_str;

	if (strncmp(option_name, "macaddr", 7) == 0)
		len = MAC_ADDR_LEN - 1;
	else
		len = MAX_URL_LENGTH - 1;

	uci_foreach_element(&s->options, element) {
		option = uci_to_option(element);

		if (strcmp(option->e.name, option_name) == 0) {
			switch (option->type) {
			case UCI_TYPE_LIST:
				// handle list available as
				// list name value1
				// list name value2
				// ....
				uci_foreach_element(&option->v.list, bl_element) {
					// for URL we need to strip the protocol if exist in the fqdn
					char *tmp = bl_element->name;
					if (strncmp(tmp, "http://", 7) == 0)
						tmp = tmp + 7;
					else if (strncmp(tmp, "https://", 8) == 0)
						tmp = tmp + 8;

					strncpy(value[i], tmp, len);
					i++;
				}
				break;
			case UCI_TYPE_STRING:
				// handle list available as
				// option name value1 value2 value3 ...
				if (!option->v.string || (option->v.string)[0] == '\0')
					break;

				local_str = strdup(option->v.string);
				tok = strtok_r(local_str, " ", &tok_r);
				while (tok != NULL) {
					// for URL we need to strip the protocol if exist in the fqdn
					char *tmp = tok;
					if (strncmp(tmp, "http://", 7) == 0)
						tmp = tmp + 7;
					else if (strncmp(tmp, "https://", 8) == 0)
						tmp = tmp + 8;

					strncpy(value[i], tmp, len);
					i++;
					tok = strtok_r(NULL, " ", &tok_r);
				}
				free(local_str);
				break;
			default:
				break;
			}
		}
	}
}

// This function reads the profile sections in the uci file and stores the info
// in global variable url_profile.
// @return 0 on success, -1 otherwise
static int read_profile_data(void)
{
	struct uci_element *e = NULL;
	struct uci_section *s = NULL;

	// iterate to get the number of profile sections
	uci_foreach_element(&url_pkg->sections, e) {
		s = uci_to_section(e);

		// read only profile section, skip the rest
		if (strcmp(s->type, "profile") != 0)
			continue;

		profile_cnt++;
	}

	if (profile_cnt == 0) {
		syslog(LOG_INFO, "Profile section not present in config");
		return 0;
	}

	url_profile = (profile *)calloc(profile_cnt, sizeof(profile));

	unsigned int i = 0;

	uci_foreach_element(&url_pkg->sections, e) {
		s = uci_to_section(e);

		// read only profile section, skip the rest
		if (strcmp(s->type, "profile") != 0)
			continue;

		get_profile_name(i, s);

		if ((url_profile + i)->name[0] == '\0')
			continue;

		create_url_list(i, s);
		populate_string_from_list(s, "blacklist_url", (url_profile + i)->blacklist_url);
		populate_string_from_list(s, "whitelist_url", (url_profile + i)->whitelist_url);
		i++;
	}

	return 0;
}

// This function allocated memory to url_filter->mac_list and It also stores the
// value of the number of mac address stored in this profile that is url_filter->mac_cnt
// @param i is the index of glabal array url_filter, this url_filter
//          for which the memory allocation will be done
// @param s is the pointer to the uci structure to be used to extract number of
//          maclist entries in the profile section
// @return nothing
static void create_mac_list(size_t i, struct uci_section *s)
{
	unsigned int maclist_count = get_list_entries_count(s, "macaddr");

	// Allocate memory to array of strings, this will define the number
	// of entries in the variable, so no of array elements
	(url_filter + i)->mac_list = (char **)calloc(maclist_count, sizeof(char *));

	(url_filter + i)->mac_cnt = maclist_count;

	int j;

	for (j = 0; j < maclist_count; j++)
		(url_filter + i)->mac_list[j] = (char *)calloc(MAC_ADDR_LEN, sizeof(char));

}

// This function links the filter to the profile
// @param i is index of filter instance
// @param s is the pointer to the uci section
// return nothing
static void update_profile_pointer(size_t i, struct uci_section *s)
{
	struct uci_element *element = NULL;

	uci_foreach_element(&s->options, element) {
	struct uci_option *option = (uci_to_option(element));
		if (strcmp(option->e.name, "profile") != 0)
			continue;

		int j;

		for (j = 0; j < profile_cnt; j++) {
			if (strcmp(option->v.string, (url_profile + j)->name) == 0) {
				(url_filter + i)->p = &url_profile[j];
				break;
			}
		}

		// profile is an option so there will be only 1 per filter section,
		// hence, nothing to do beyond this point, just break
		break;
	}
}

// This function coverts day of week to index, Sunday being 0, this is in line
// with the tm structure in the time.h file and eases comparison
// @param i is index of url_filter to populate
// @param day is day of the week
// @return nothing
static void convert_day_to_index(size_t i, char *day)
{
	if (strcasecmp(day, "Sunday") == 0)
		(url_filter + i)->day[0] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Monday") == 0)
		(url_filter + i)->day[1] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Tuesday") == 0)
		(url_filter + i)->day[2] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Wednesday") == 0)
		(url_filter + i)->day[3] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Thursday") == 0)
		(url_filter + i)->day[4] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Friday") == 0)
		(url_filter + i)->day[5] = FILTER_ACTIVE;
	else if (strcasecmp(day, "Saturday") == 0)
		(url_filter + i)->day[6] = FILTER_ACTIVE;

}

// This fucntion parses the list of days in the uci section s and sets the days
// on which url filtering is enabled in the array.
// @param i is index of the url_filter to update
// @param s is pointer to uci section
// @return nothing
static void update_active_days(size_t i, struct uci_section *s)
{
	struct uci_element *element = NULL;
	struct uci_option *option = NULL;
	struct uci_element *bl_element = NULL;
	char *local_str;
	char *tok, *tok_r;

	int j;

	uci_foreach_element(&s->options, element) {
		option = (uci_to_option(element));
		if (strcmp(option->e.name, "day") != 0)
			continue;

		for (j = 0; j < 7; j++)
			(url_filter + i)->day[j] = FILTER_INACTIVE;

		switch (option->type) {
		case UCI_TYPE_LIST:
			// handle list available as
			// list name value1
			// list name value2
			// ....
			uci_foreach_element(&option->v.list, bl_element) {
				convert_day_to_index(i, bl_element->name);

			}
			break;
		case UCI_TYPE_STRING:
			// handle list available as
			// option name value1 value2 value3 ...
			if (!option->v.string || (option->v.string)[0] == '\0')
				break;

			local_str = strdup(option->v.string);
			tok = strtok_r(local_str, " ", &tok_r);
			while (tok != NULL) {
				convert_day_to_index(i, tok);
				tok = strtok_r(NULL, " ", &tok_r);
			}
			free(local_str);
			break;
		default:
			break;
		}
	}
}

// This function reads the config filter section of the uci file and stores the info
// in filter structure.
static int read_filter_data(void)
{
	struct uci_element *e = NULL;
	struct uci_section *s = NULL;

	struct uci_element *element = NULL;
	struct uci_option *option = NULL;

	// iterate to get the number of filter sections
	uci_foreach_element(&url_pkg->sections, e) {
		s = uci_to_section(e);

		// read only filter section, skip the rest
		if (strcmp(s->type, "filter") != 0)
			continue;

		filter_cnt++;
	}

	if (filter_cnt == 0) {
		syslog(LOG_INFO, "Filter section not present in config");
		return 0;
	}

	url_filter = (filter *)calloc(filter_cnt, sizeof(filter));

	unsigned int i = 0;

	uci_foreach_element(&url_pkg->sections, e) {
		s = uci_to_section(e);

		// read only filter section, skip the rest
		if (strcmp(s->type, "filter") != 0)
			continue;

		create_mac_list(i, s);
		populate_string_from_list(s, "macaddr", (url_filter + i)->mac_list);
		update_profile_pointer(i, s);
		update_active_days(i, s);

		uci_foreach_element(&s->options, element) {
			option = uci_to_option(element);
			if (strcmp(option->e.name, "enable") == 0) {
				if (strcmp(option->v.string, "1") == 0)
					(url_filter + i)->enable = true;
				else
					(url_filter + i)->enable = false;
			}

			if (strcmp(option->e.name, "start_time") == 0)
				strncpy((url_filter + i)->start_time, option->v.string, sizeof((url_filter + i)->start_time));

			if (strcmp(option->e.name, "duration") == 0)
				(url_filter + i)->duration = (unsigned int)strtoul(option->v.string, NULL, 10);
		}

		i++;
	}

	return 0;
}

// This function reads the config global section of the uci file and stores the info
// in global structure.
static int read_global_data(void)
{
	struct uci_element *e = NULL;
	struct uci_section *s = NULL;

	struct uci_element *element = NULL;
	struct uci_option *option = NULL;

	uci_foreach_element(&url_pkg->sections, e) {
		s = uci_to_section(e);

		// read only global section, skip the rest
		if (strcmp(s->type, "globals") != 0)
			continue;

		bool is_glb_url_read = false;

		uci_foreach_element(&s->options, element) {
			option = uci_to_option(element);
			if (strcmp(option->e.name, "enable") == 0) {
				if (strcmp(option->v.string, "1") == 0)
					url_global.enable_filtering = true;
				else
					url_global.enable_filtering = false;
			}

			if (strcmp(option->e.name, "global_blacklist") == 0) {
				if (strcmp(option->v.string, "1") == 0)
					url_global.enable_global_blacklisting = true;
				else
					url_global.enable_global_blacklisting = false;
			}

			if ((strcmp(option->e.name, "blacklist_url") == 0) && (url_global.enable_global_blacklisting == true)) {
				if (is_glb_url_read == true)
					continue;

				unsigned int glb_url_cnt = get_list_entries_count(s, "blacklist_url");

				url_global.blacklist_url = (char **)calloc(glb_url_cnt, sizeof(char *));
				url_global.blacklist_url_cnt = glb_url_cnt;

				int j;

				for (j = 0; j < glb_url_cnt; j++)
					url_global.blacklist_url[j] = (char *)calloc(MAX_URL_LENGTH, sizeof(char));

				populate_string_from_list(s, "blacklist_url", url_global.blacklist_url);
				is_glb_url_read = true;
			}

		}
		// only 1 global section in uci so no point to loop further
		break;
	}

	return 0;
}

// This function populates the blacklisted urls to be used when blocking https packets
static int init_dns_cache(void)
{
	dns_cache_cnt = url_global.blacklist_url_cnt;
	int i;

	for (i = 0; i < profile_cnt; i++)
		dns_cache_cnt += (url_profile + i)->blacklist_url_cnt;

	global_dns_cache = (dns_cache *)calloc(dns_cache_cnt, sizeof(dns_cache));

	int j = 0;

	for (i = 0; i < url_global.blacklist_url_cnt; i++) {
		strncpy((global_dns_cache + j)->url, url_global.blacklist_url[i], sizeof((global_dns_cache + j)->url));
		j++;
	}

	int k = 0;

	for (k = 0; k < profile_cnt; k++) {
		for (i = 0; i < (url_profile + k)->blacklist_url_cnt; i++) {
			strncpy((global_dns_cache + j)->url, (url_profile + k)->blacklist_url[i], sizeof((global_dns_cache + j)->url));
			j++;
		}
	}

	return 0;
}

// this function reads the uci config from /etc/config/urlfiter file and store the
// information in daemon's data structures
int read_config(void)
{
	url_ctx = uci_alloc_context();
	if (!url_ctx) {
		syslog(LOG_ERR, "failed to allocate uci context");
		return -1;
	}

	if (uci_load(url_ctx, "urlfilter", &url_pkg)) {
		syslog(LOG_ERR, "failed to load uci file");
		return -1;
	}

	if (read_global_data() != 0) {
		syslog(LOG_ERR, "error in reading global section");
		return -1;
	}

	// If url filtering is disabled, no point of proceeding beyond this point
	if (url_global.enable_filtering == false) {
		syslog(LOG_INFO, "url filtering is disabled");
		return -1;
	}

	if (read_profile_data() != 0) {
		syslog(LOG_ERR, "error in reading profile section");
		return -1;
	}

	if (read_filter_data() != 0) {
		syslog(LOG_ERR, "error in reading filter section");
		return -1;
	}

	if (init_dns_cache() != 0) {
		syslog(LOG_ERR, "error in initialising dns cache");
		return -1;
	}

	return 0;
}

